# encoding: utf-8
require "logstash/outputs/base"
require "logstash/namespace"
require "logstash/json"
require "uri"
require "logstash/plugin_mixins/http_client"

class LogStash::Outputs::Faye < LogStash::Outputs::Base
  include LogStash::PluginMixins::HttpClient

  # This output lets you send events to a
  # generic HTTP(S) endpoint
  #
  # This output will execute up to 'pool_max' requests in parallel for performance.
  # Consider this when tuning this plugin for performance.
  #
  # Additionally, note that when parallel execution is used strict ordering of events is not
  # guaranteed!
  #
  # Beware, this gem does not yet support codecs. Please use the 'format' option for now.

  config_name "faye"

  # Faye-channel on to be sent on
  config :channel, :validate => :string

  # Faye-security-token
  config :faye_token, :validate => :string

  # URL to use
  config :url, :validate => :string, :default => "http://localhost:9292/faye"

  # This lets you choose the structure and parts of the event that are sent.
  #
  #
  # For example:
  # [source,ruby]
  #    mapping => {"foo" => "%{host}" 
  #               "bar" => "%{type}"}
  config :mapping, :validate => :hash

  def register
    @http_method = :post
    @content_type = "application/json"

    # We count outstanding requests with this queue
    # This queue tracks the requests to create backpressure
    # When this queue is empty no new requests may be sent,
    # tokens must be added back by the client on success
    @request_tokens = SizedQueue.new(@pool_max)
    @pool_max.times {|t| @request_tokens << true }

    @requests = Array.new
  end # def register

  def multi_receive(events)
    events.each {|event| receive(event, :parallel)}
    client.execute!
  end

  # Once we no longer need to support Logstash < 2.2 (pre-ng-pipeline)
  # We don't need to handle :background style requests
  #
  # We use :background style requests for Logstash < 2.2 because before the microbatching
  # pipeline performance is greatly improved by having some degree of async behavior.
  #
  # In Logstash 2.2 and after things are much simpler, we just run each batch in parallel
  # This will make performance much easier to reason about, and more importantly let us guarantee
  # that if `multi_receive` returns all items have been sent.
  def receive(event, async_type=:background)

    if @faye_token
      body = LogStash::Json.dump({:channel => @channel, :data => map_event(event), :ext => {:auth_token => @faye_token}})
    else
      body = LogStash::Json.dump({:channel => @channel, :data => map_event(event)})
    end

    # Block waiting for a token
    token = @request_tokens.pop if async_type == :background

    # Send the request
    url = event.sprintf(@url)
    headers = event_headers(event)

    # Create an async request
    request = client.send(async_type).send(@http_method, url, :body => body, :headers => headers)

    request.on_complete do
      # Make sure we return the token to the pool
      @request_tokens << token  if async_type == :background
    end

    request.on_success do |response|
      if response.code < 200 || response.code > 299
        log_failure(
          "Encountered non-200 HTTP code #{response.code}",
          :response_code => response.code,
          :url => url,
          :event => event.to_hash)
      end
    end

    request.on_failure do |exception|
      log_failure("Could not fetch URL",
                  :url => url,
                  :method => @http_method,
                  :body => faye_body.to_json,
                  :headers => headers,
                  :message => exception.message,
                  :class => exception.class.name,
                  :backtrace => exception.backtrace
      )
    end

    request.call if async_type == :background
  end

  def close
    client.close
  end

  private

  # This is split into a separate method mostly to help testing
  def log_failure(message, opts)
    @logger.error("[HTTP Output Failure] #{message}", opts)
  end

  # Manticore doesn't provide a way to attach handlers to background or async requests well
  # It wants you to use futures. The #async method kinda works but expects single thread batches
  # and background only returns futures.
  # Proposed fix to manticore here: https://github.com/cheald/manticore/issues/32
  def request_async_background(request)
    @method ||= client.executor.java_method(:submit, [java.util.concurrent.Callable.java_class])
    @method.call(request)
  end

  def map_event(event)
    if @mapping
      @mapping.reduce({}) do |acc,kv|
        k,v = kv
        acc[k] = event.sprintf(v)
        acc
      end
    else
      event.to_hash
    end
  end

  def event_headers(event)
    headers = custom_headers(event) || {}
    headers["Content-Type"] = @content_type
    headers
  end

  def custom_headers(event)
    return nil unless @headers

    @headers.reduce({}) do |acc,kv|
      k,v = kv
      acc[k] = event.sprintf(v)
      acc
    end
  end
end
